#include "..\image.hpp"

extern "C" int __cdecl _purecall(void) { return 0; }

BYTE mask[] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
BGR  black  = { 0, 0, 0 };
BGR  white  = { 0xff, 0xff, 0xff };

#define N0    8
#define N1    191
#define B0    32
#define B1    255

BGR palette[] =
{
    {0,0,0}, {N1,N0,N0}, {N0,N0,N1}, {N1,N0,N1}, {N0,N1,N0}, {N1,N1,N0}, {N0,N1,N1}, {N1,N1,N1},
    {0,0,0}, {B1,B0,B0}, {B0,B0,B1}, {B1,B0,B1}, {B0,B1,B0}, {B1,B1,B0}, {B0,B1,B1}, {B1,B1,B1}
};


bool checkSignature(const BYTE* data, const BYTE signature[])
{
  for(size_t i = 0; i < sizeof(signature); ++i)
    if(data[i] != signature[i]) return false;
  return true;
}

bool scr(Image* im, const BYTE* data, bool secondStep = false, bool skipAttr = false)
{
  Page* pg = im->addPage();

  bool flash = false;
  if(pg)
  {
    if(secondStep) im->animate = true;

    for(size_t y = 0; y < 192; ++y)
    {
      const BYTE* pixelLine = data + 32 * ((y & 0xc0) | ((y << 3) & 0x38) | ((y >> 3) & 0x07));
      const BYTE* attr      = data + 6144 + ((y & 0xf8) << 2);

      for(size_t x = 0; x < 32; ++x)
      {
        BYTE paper = (attr[x] >> 3  ) & 0x0f;
        BYTE ink   = (attr[x] & 0x07) | (paper & 0x08);
        BYTE pix   = pixelLine[x];

        if(skipAttr) { ink = 0; paper = 7; }

        if((attr[x] & 0x80) && paper != ink) flash = true;

        for(size_t bit = 0; bit < 8; ++bit)
        {
          size_t i  = 256*y + 8*x + bit;
          size_t ci = (pix & mask[bit]) ? ink : paper;

          if(secondStep && (attr[x] & 0x80))
              ci = (pix & mask[bit]) ? paper : ink;

          pg->pix[i] = palette[ci];
        }
      }
    }
  }
  return flash;
}

Image* scr2image(const BYTE* data, bool hasAttr)
{
  Image* im = new Image();

  BYTE buf[6912];
  const BYTE* p = data;

  if(!hasAttr)
  {
    for(int i = 0; i < 6144; ++i) buf[i] = data[i];
    for(int i = 6144; i < 6912; ++i) buf[i] = 7;
    p = buf;
  }

  if(scr(im, p))
    scr(im, p, true);
  else
  {
    if(hasAttr)
      scr(im, p, false, true);
    else
    {
      // second page: screen in "sprite" format
      Page* pg = im->addPage();
      if(pg)
      {
        for(size_t y = 0; y < 192; ++y)
        {
          const BYTE* pixelLine = data + 32 * y;
          for(size_t x = 0; x < 32; ++x)
          {
            BYTE pix   = pixelLine[x];
            for(size_t bit = 0; bit < 8; ++bit)
            {
              size_t i  = 256*y + 8*x + bit;
              size_t ci = (pix & mask[bit]) ? 0 : 7;

              pg->pix[i] = palette[ci];
            }
          }
        }
      }
    }
  }
  return im;
}


class HrustStream
{
  private:
    const BYTE* base;
    const BYTE* p;
    int   idx;
    int   len;
    bool  eof;
    WORD  bits;
  public:
    HrustStream( const BYTE* from, int blockSize )
    {
      base = p = from;

      len = blockSize;
      idx = 0;
      eof = false;

      bits  = getByte();
      bits += 256*getByte();
    }

    BYTE getByte( void )
    {
      if( p - base == len ) { eof = true; return 0; }
      return *p++;
    }

    BYTE getBit()
    {
      WORD mask[]  = { 0x8000, 0x4000, 0x2000, 0x1000, 0x800, 0x400, 0x200, 0x100, 0x80, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x1 };

      BYTE bit = ( bits & mask[idx] ) ? 1 : 0;
      if( idx == 15 )
      {
        bits  = getByte();
        bits += 256*getByte();
      }

      idx = (idx + 1) % 16;
      return bit;
    }

    BYTE getBits( int n )
    {
      BYTE r = 0;
      do { r = 2*r + getBit(); } while( --n );
      return r;
    }

    bool error( void ) { return eof; }
};


void dehrust(const BYTE* source, BYTE* destination)
{
  const BYTE *from = source;

  if( from[0] != 'H' || from[1] != 'R' ) from += 0x103;

  HrustStream s( from + 12, from[4] + 256*from[5] );

  BYTE* to = destination;

  *to++ = s.getByte();

  BYTE noBits = 2;
  BYTE mask[] = { 0, 0, 0xfc, 0xf8, 0xf0, 0xe0, 0xc0, 0x80, 0 };

  while( !s.error() )
  {
    while( s.getBit() ) *to++ = s.getByte();

    WORD len = 0;
    BYTE bb  = 0;
    do
    {
      bb = s.getBits( 2 );
      len += bb;
    } while( bb == 0x03 && len != 0x0f );

    short offset = 0;

    if( len == 0 )
    {
      offset = 0xfff8 + s.getBits( 3 );
      *to++ = to[offset];
      continue;
    }

    if( len == 1 )
    {
      BYTE code = s.getBits(2);

      if( code == 2 )
      {
        BYTE b = s.getByte();
        if( b >= 0xe0 )
        {
          b <<= 1; ++b; // rlca
          b ^= 2;       // xor c

          if( b == 0xff ) { ++noBits; continue; }

          offset = 0xff00 + b - 0x0f;

          *to++ = to[offset];
          *to++ = s.getByte();
          *to++ = to[offset];
          continue;
        }
        offset = 0xff00 + b;
      }

      if( code == 0 || code == 1 )
      {
        offset = s.getByte();
        offset += 256*(code ? 0xfe : 0xfd );
      }
      if( code == 3 ) offset = 0xffe0 + s.getBits( 5 );

      for( BYTE i = 0; i < 2; ++i ) *to++ = to[offset];
      continue;
    }

    if( len == 3 )
    {
      if( s.getBit() )
      {
        offset = 0xfff0 + s.getBits( 4 );
        *to++ = to[offset];
        *to++ = s.getByte();
        *to++ = to[offset];
        continue;
      }

      if( s.getBit() )
      {
        BYTE noBytes = 6 + s.getBits(4);
        for( BYTE i = 0; i < 2*noBytes; ++i ) *to++ = s.getByte();
        continue;
      }

      len = s.getBits( 7 );
      if( len == 0x0f ) break; // EOF
      if( len <  0x0f ) len = 256*len + s.getByte();
    }

    if( len == 2 ) ++len;

    BYTE code = s.getBits( 2 );

    if( code == 1 )
    {
      BYTE b = s.getByte();

      if( b >= 0xe0 )
      {
        if( len > 3 ) return;

        b <<= 1; ++b; // rlca
        b ^= 3;       // xor c

        offset = 0xff00 + b - 0x0f;

        *to++ = to[offset];
        *to++ = s.getByte();
        *to++ = to[offset];
        continue;
      }
      offset = 0xff00 + b;
    }

    if( code == 0 ) offset = 0xfe00 + s.getByte();
    if( code == 2 ) offset = 0xffe0 + s.getBits( 5 );
    if( code == 3 )
    {
      offset  = 256*( mask[noBits] + s.getBits(noBits) );
      offset += s.getByte();
    }

    for( WORD i = 0; i < len; ++i ) *to++ = to[offset];
  }

  for( int i = 0; i < 6; ++i ) *to++ = from[6+i];
}
